@@ -550,7 +550,9 @@ class Parser {
         for (int j = start + 1; j < i; j++) {
           Regexp subMax = array[s + max], subJ = array[s + j];
           if (subMax.op.ordinal() < subJ.op.ordinal()
-              || (subMax.op == subJ.op && subMax.runes.length < subJ.runes.length)) {
+              || (subMax.op == subJ.op
+                  && (subMax.runes != null ? subMax.runes.length : 0)
+                      < (subJ.runes != null ? subJ.runes.length : 0))) {
             max = j;
           }
         }
@@ -782,7 +784,7 @@ class Parser {
   }
 
   /**
-   * Parse regular expression pattern {@var pattern} with mode flags {@var flags}.
+   * Parse regular expression pattern {@code pattern} with mode flags {@code flags}.
    */
   static Regexp parse(String pattern, int flags) throws PatternSyntaxException {
     return new Parser(pattern, flags).parseInternal();
@@ -922,8 +924,10 @@ class Parser {
                     }
                     t.skipString(lit);
                     t.skipString("\\E");
-                    for (int j = 0; j < lit.length(); j++) {
-                      literal(lit.charAt(j));
+                    for (int j = 0; j < lit.length(); ) {
+                      int codepoint = lit.codePointAt(j);
+                      literal(codepoint);
+                      j += Character.charCount(codepoint);
                     }
                     break bigswitch;
                   }
