@@ -249,6 +249,7 @@ public class FindTest {
         35,
         35,
         36),
+    new Test("(|a)*", "aa", 3, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2),
   };
 
   @Parameters
@@ -481,7 +482,9 @@ public class FindTest {
     for (int k = 0; k < GoTestUtils.len(result); k++) {
       if (submatches[k * 2] == -1) {
         if (result[k] != null) {
-          fail(String.format("%s %d: expected null got %s: %s", testName, n, result, test));
+          fail(
+              String.format(
+                  "%s %d: expected null got %s: %s", testName, n, Arrays.toString(result), test));
         }
         continue;
       }
@@ -538,7 +541,14 @@ public class FindTest {
       System.err.println(testName + "  " + test + " " + n + " " + k + " ");
       String expect = test.submatchString(n, k / 2);
       if (!expect.equals(result[k / 2])) {
-        fail(String.format("%s %d: expected %s got %s: %s", testName, n, expect, result, test));
+        fail(
+            String.format(
+                "%s %d: expected %s got %s: %s",
+                testName,
+                n,
+                expect,
+                Arrays.toString(result),
+                test));
       }
     }
   }
